"""API Schemas package for reusable and complex schema definitions."""

from .common import (
    BaseResponse,
    PaginationParams,
    PaginatedResponse,
    StatusUpdateRequest,
    IDResponse,
    BulkResponse,
    BaseEventCreateRequest,
    BaseEventResponse,
)
from .alignment import (
    AlignmentProcessingProgress,
    AlignmentProcessingStatus,
    AlignmentResponse,
    AlignmentStatus,
    AlignmentWorkflowStatus,
    AlignmentCreateRequest,
    AlignmentCloseRequest,
    AlignmentOpenRequest,
    AlignmentUpdateRequest,
    AlignmentMetrics,
    AlignmentTypeResponse,
    AlignmentTypeCreateRequest,
    AlignmentExecuteRequest,
    AlignmentExecuteResponse,
    AlignmentEventResponse,
)
from .bucket import (
    BucketResponse,
    BucketDetailResponse,
    BucketListItemResponse,
    PaginatedBucketListResponse,
    PaginatedBucketTaskGroupSummaryResponse,
    BucketCreateRequest,
    BucketStatusUpdateRequest,
    BucketEventCreateRequest,
    BucketEventResponse,
    BucketMetrics,
    BucketTaskSnapshot,
    BucketSummaryResponse,
)
from .command import (
    CommandResponse,
    CommandGroupResponse,
    CommandGroupCreateRequest,
    CommandGroupUpdateRequest,
    CommandGroupStatusUpdateRequest,
    CommandCreateRequest,
    ApproveCommandRequest,
    PendingCommandRequest,
    RejectCommandRequest,
    CommandGroupWithCommandsResponse,
    CommandFindingsMapCreateRequest,
    UpdateCommandFindingsMappingRequest,
    CommandFindingsMapResponse,
    CommandGroupEventCreateRequest,
    CommandEventCreateRequest,
    CommandEventResponse,
    CommandGroupEventResponse,
    ProjectionEventSchema,
    QueryReferenceOptionsParams,
    QueryCommandsImpactingPayloadParams,
    QueryDisplayOptionsFromItemsParams,
    QueryDisplayAttributesResult,
    CommandImpactingPayloadResponse,
    CommandGroupStatus,
    CommandStatus,
    CommandType,
)
from .findings import (
    FindingResponse,
    FindingListItemResponse,
    PaginatedFindingListResponse,
    FindingCreateRequest,
    AddFindingRequest,
    RetireFindingRequest,
    UpdateFindingRequest,
    FindingStatsResponse,
    FindingMatchCreateRequest,
    FindingMatchResponse,
    FindingMatchQueryCreateRequest,
    FindingMatchQueryUpdateRequest,
    FindingMatchQueryResponse,
    QueryProviderResponse,
    PatchFindingAdditionalDataRequest,
    FindingEventResponse,
)
from .post_processing import (
    MapFindingsRequest,
    SiblingKeyConfig,
    AlignmentGraphExpandRequest,
    AlignmentGraphStatusResponse,
    AlignmentGraphOperationResponse,
    AlignmentGraphStartProcessingRequest,
    AlignmentGraphStartProcessingResponse,
    BucketFindingsResponse,
    BucketCommandsResponse,
    BucketCommandsDropResponse,
    AlignmentFindingsResponse,
    AlignmentCommandsResponse,
    AlignmentCommandsDropResponse,
    AlignmentFindingsSummaryReportRequest,
    AlignmentFindingsSummaryReportResponse,
    AlignmentArtifactsArchiveResponse,
    PostProcessingJobResponse,
    PostProcessingJobStatus,
)
from .impact_assessment import (
    ImpactAssessmentResponse,
    ImpactAssessmentWithDependenciesResponse,
    DependencySummaryEntry,
    DependencyItemTypeCountEntry,
    DependencyNodeResponse,
    DependencyEdgeResponse,
    DependencyGraphResponse,
    FindingDependencyGraphRequest,
    CommandNodeResponse,
    CommandEdgeResponse,
    CommandGraphResponse,
    DependencyNodeExpansionState,
    DependencyGraphPageInfo,
    DependencyGraphExpandNodeRequest,
    DependencyGraphExpandRequest,
    DependencyGraphExpandContext,
    DependencyGraphExpandDelta,
    DependencyGraphExpandResponse,
)
from .registration import (
    ProcessorEndpointsSchema,
    ProcessorRegistrationResponse,
    ProcessorRegistrationUpdateRequest,
    ProcessorRegistrationUpsertRequest,
)

__all__ = [
    # Common
    "BaseResponse",
    "PaginationParams",
    "PaginatedResponse",
    "StatusUpdateRequest",
    "IDResponse",
    "BulkResponse",
    "BaseEventCreateRequest",
    "BaseEventResponse",
    # Alignment
    "AlignmentResponse",
    "AlignmentProcessingProgress",
    "AlignmentProcessingStatus",
    "AlignmentStatus",
    "AlignmentWorkflowStatus",
    "AlignmentCreateRequest",
    "AlignmentCloseRequest",
    "AlignmentOpenRequest",
    "AlignmentUpdateRequest",
    "AlignmentMetrics",
    "AlignmentTypeResponse",
    "AlignmentTypeCreateRequest",
    "AlignmentExecuteRequest",
    "AlignmentExecuteResponse",
    "AlignmentEventResponse",
    # Bucket
    "BucketResponse",
    "BucketDetailResponse",
    "BucketListItemResponse",
    "PaginatedBucketListResponse",
    "PaginatedBucketTaskGroupSummaryResponse",
    "BucketCreateRequest",
    "BucketStatusUpdateRequest",
    "BucketEventCreateRequest",
    "BucketEventResponse",
    "BucketMetrics",
    "BucketTaskSnapshot",
    "BucketSummaryResponse",
    # Command
    "CommandResponse",
    "CommandGroupResponse",
    "CommandGroupCreateRequest",
    "CommandGroupUpdateRequest",
    "CommandGroupStatusUpdateRequest",
    "CommandCreateRequest",
    "ApproveCommandRequest",
    "PendingCommandRequest",
    "RejectCommandRequest",
    "CommandGroupWithCommandsResponse",
    "CommandFindingsMapCreateRequest",
    "UpdateCommandFindingsMappingRequest",
    "CommandFindingsMapResponse",
    "CommandGroupEventCreateRequest",
    "CommandEventCreateRequest",
    "CommandEventResponse",
    "CommandGroupEventResponse",
    "ProjectionEventSchema",
    "QueryReferenceOptionsParams",
    "QueryCommandsImpactingPayloadParams",
    "QueryDisplayOptionsFromItemsParams",
    "QueryDisplayAttributesResult",
    "CommandImpactingPayloadResponse",
    "CommandGroupStatus",
    "CommandStatus",
    "CommandType",
    # Finding
    "FindingResponse",
    "FindingListItemResponse",
    "PaginatedFindingListResponse",
    "FindingCreateRequest",
    "AddFindingRequest",
    "RetireFindingRequest",
    "UpdateFindingRequest",
    "FindingStatsResponse",
    "FindingMatchCreateRequest",
    "FindingMatchResponse",
    "FindingMatchQueryCreateRequest",
    "FindingMatchQueryUpdateRequest",
    "FindingMatchQueryResponse",
    "QueryProviderResponse",
    "PatchFindingAdditionalDataRequest",
    "FindingEventResponse",
    # Post-Processing
    "MapFindingsRequest",
    "SiblingKeyConfig",
    "AlignmentGraphExpandRequest",
    "AlignmentGraphStatusResponse",
    "AlignmentGraphOperationResponse",
    "AlignmentGraphStartProcessingRequest",
    "AlignmentGraphStartProcessingResponse",
    "BucketFindingsResponse",
    "BucketCommandsResponse",
    "BucketCommandsDropResponse",
    "AlignmentFindingsResponse",
    "AlignmentCommandsResponse",
    "AlignmentCommandsDropResponse",
    "AlignmentFindingsSummaryReportRequest",
    "AlignmentFindingsSummaryReportResponse",
    "AlignmentArtifactsArchiveResponse",
    "PostProcessingJobResponse",
    "PostProcessingJobStatus",
    # Impact Assessment
    "ImpactAssessmentResponse",
    "DependencySummaryEntry",
    "DependencyItemTypeCountEntry",
    "ImpactAssessmentWithDependenciesResponse",
    "DependencyNodeResponse",
    "DependencyEdgeResponse",
    "DependencyGraphResponse",
    "FindingDependencyGraphRequest",
    "CommandNodeResponse",
    "CommandEdgeResponse",
    "CommandGraphResponse",
    "DependencyNodeExpansionState",
    "DependencyGraphPageInfo",
    "DependencyGraphExpandNodeRequest",
    "DependencyGraphExpandRequest",
    "DependencyGraphExpandContext",
    "DependencyGraphExpandDelta",
    "DependencyGraphExpandResponse",
    # Processor registration
    "ProcessorEndpointsSchema",
    "ProcessorRegistrationResponse",
    "ProcessorRegistrationUpdateRequest",
    "ProcessorRegistrationUpsertRequest",
]



"""Impact Assessment-related schemas."""

import datetime
from typing import Any
from uuid import UUID
from pydantic import AliasChoices, Field
from .common import BaseResponse, BaseSchema


# ---------------------------------------------------------------------------
# Graph expansion / query schemas
# ---------------------------------------------------------------------------


class DependencyNodeExpansionState(BaseSchema):
    """Metadata about unexplored connections for a graph node."""

    has_incoming: bool = Field(
        ..., description="Node has any incoming edges in the stored graph"
    )
    has_outgoing: bool = Field(
        ..., description="Node has any outgoing edges in the stored graph"
    )
    incoming_count: int = Field(
        ..., description="Total incoming edges in the stored graph"
    )
    outgoing_count: int = Field(
        ..., description="Total outgoing edges in the stored graph"
    )
    is_incoming_truncated: bool = Field(
        ..., description="More incoming edges exist beyond what was loaded"
    )
    is_outgoing_truncated: bool = Field(
        ..., description="More outgoing edges exist beyond what was loaded"
    )


class DependencyGraphPageInfo(BaseSchema):
    """Pagination / limit metadata for a dependency graph query."""

    applied_limit: int | None = Field(
        default=None, description="Relationship limit that was applied"
    )
    data_limit_mode: str = Field(default="same_parent", description="Limit mode used")
    was_truncated: bool = Field(
        default=False, description="Graph was truncated by any limit"
    )
    omitted_by_limit: bool = Field(
        default=False, description="Edges were omitted due to relationship_limit"
    )


# ---------------------------------------------------------------------------
# Graph expand request / response schemas
# ---------------------------------------------------------------------------


class DependencyGraphExpandNodeRequest(BaseSchema):
    """Request to expand a single node in the dependency graph."""

    item_id: str = Field(..., description="Item UUID of the node to expand")
    direction: str = Field(
        default="both",
        description="Edge direction to expand: 'incoming', 'outgoing', or 'both'",
    )
    limit: int | None = Field(default=None, gt=0, description="Max edges to return")
    cursor: str | None = Field(
        default=None, description="Pagination cursor from a previous expand"
    )


class DependencyGraphExpandRequest(BaseSchema):
    """Request body for the graph expand endpoint."""

    expand: list[DependencyGraphExpandNodeRequest] = Field(
        default_factory=list, description="Nodes to expand"
    )
    known_node_ids: list[str] = Field(
        default_factory=list,
        description="Node item_ids already held by the client (excluded from delta)",
    )
    known_edge_ids: list[str] = Field(
        default_factory=list,
        description="Edge IDs already held by the client (format 'source:target')",
    )


class DependencyGraphExpandContext(BaseSchema):
    """Context returned alongside an expand delta."""

    alignment_id: str = Field(..., description="Alignment the graph belongs to")
    expanded_node_ids: list[str] = Field(
        default_factory=list, description="Node IDs that were expanded"
    )
    next_cursors: dict[str, str | None] = Field(
        default_factory=dict, description="Next pagination cursor per expanded node ID"
    )


class DependencyGraphExpandDelta(BaseSchema):
    """Incremental graph delta returned by the expand endpoint."""

    added_nodes: list["DependencyNodeResponse"] = Field(
        default_factory=list, description="New nodes not previously known to the client"
    )
    added_edges: list["DependencyEdgeResponse"] = Field(
        default_factory=list, description="New edges not previously known to the client"
    )


class DependencyGraphExpandResponse(BaseSchema):
    """Response from the graph expand endpoint."""

    context: DependencyGraphExpandContext
    delta: DependencyGraphExpandDelta


class ImpactAssessmentResponse(BaseResponse):
    """Response schema for impact assessment data."""

    id: int = Field(..., description="Unique identifier for the assessment")
    finding_id: UUID = Field(..., description="ID of the associated finding")
    created_at: datetime.datetime = Field(
        ...,
        description="Creation timestamp",
        validation_alias=AliasChoices("timestamp", "createdAt", "created_at"),
    )


# Dependency Graph schemas
class DependencyNodeResponse(BaseSchema):
    """Response schema for a dependency graph node."""

    item_type: str = Field(..., description="Item type UUID")
    version: str = Field(..., description="Item type version")
    item_id: str = Field(..., description="Item UUID")
    is_root: bool = Field(
        ...,
        description="Whether this node is a root node directly associated with a finding",
    )
    item_data: dict[str, Any] | None = Field(
        default=None,
        description="Additional node data",
    )
    expansion_state: DependencyNodeExpansionState | None = Field(
        default=None,
        description="Expansion metadata indicating unexplored connections",
    )


class DependencyEdgeResponse(BaseSchema):
    """Response schema for a dependency graph edge."""

    source_item_id: str = Field(..., description="Source item UUID")
    target_item_id: str = Field(..., description="Target item UUID")
    status: str | None = Field(None, description="Edge status (e.g. sibling)")
    edge_id: str | None = Field(
        default=None,
        description="Stable edge ID (format 'source_item_id:target_item_id')",
    )


class DependencyGraphResponse(BaseSchema):
    """Response schema for a dependency graph."""

    nodes: list[DependencyNodeResponse] = Field(..., description="List of graph nodes")
    edges: list[DependencyEdgeResponse] = Field(..., description="List of graph edges")
    page_info: DependencyGraphPageInfo | None = Field(
        default=None,
        description="Limit and truncation metadata; present when relationship_limit is used",
    )


# Composite response schemas


class DependencySummaryEntry(BaseSchema):
    """Schema for summarizing dependencies by item type."""

    item_type: str = Field(..., description="Item type UUID")
    version: str = Field(..., description="Item type version")
    item_count: int = Field(..., description="Number of items impacted")


class DependencyItemTypeCountEntry(BaseSchema):
    """Schema for item counts grouped by item type."""

    item_type: str = Field(..., description="Item type UUID")
    item_count: int = Field(..., description="Number of projected items")


class ImpactAssessmentWithDependenciesResponse(BaseSchema):
    """Response schema for impact assessment with dependencies."""

    assessment: ImpactAssessmentResponse = Field(
        ..., description="Impact assessment data"
    )
    dependencies_summary: list[DependencySummaryEntry] | None = Field(
        default=None, description="List of dependencies"
    )
    dependencies_graph: DependencyGraphResponse | None = Field(
        None,
        description="Dependency graph showing relationships between impacted items",
    )


class FindingDependencyGraphRequest(BaseSchema):
    """Request schema for filtering finding dependency graphs."""

    finding_ids: list[UUID] | None = Field(
        None, description="Optional list of finding IDs to filter by"
    )


# Command Graph schemas
class CommandNodeResponse(BaseSchema):
    """Response schema for a command graph node."""

    command_id: str = Field(..., description="Command UUID")
    item_type_id: str = Field(..., description="Item type UUID")
    version: str = Field(..., description="Item type version")
    item_id: str = Field(..., description="Item UUID")
    data: dict[str, Any] | list[dict[str, Any]] = Field(
        default_factory=dict, description="Additional node data"
    )
    status: str | None = Field(..., description="Node status")


class CommandEdgeResponse(BaseSchema):
    """Response schema for a command graph edge."""

    source_command_id: str = Field(..., description="Source command UUID")
    source_item_id: str = Field(..., description="Source item UUID")
    target_command_id: str = Field(..., description="Target command UUID")
    target_item_id: str = Field(..., description="Target item UUID")
    status: str | None = Field(..., description="Edge status")


class CommandGraphResponse(BaseSchema):
    """Response schema for a command graph."""

    nodes: list[CommandNodeResponse] = Field(..., description="List of graph nodes")
    edges: list[CommandEdgeResponse] = Field(..., description="List of graph edges")



"""Generated catalog of Alignment Service v1 API endpoints."""

API_ENDPOINTS = [
    {"module": "alignments", "function": "delete_alignment", "method": "DELETE", "path": "/{alignment_id}"},
    {"module": "alignments", "function": "get_alignments", "method": "GET", "path": "/"},
    {"module": "alignments", "function": "get_alignment_by_name", "method": "GET", "path": "/name/{name}"},
    {"module": "alignments", "function": "get_alignments_by_type", "method": "GET", "path": "/type/{type_id}/alignments"},
    {"module": "alignments", "function": "get_alignment_types", "method": "GET", "path": "/types"},
    {"module": "alignments", "function": "get_alignment_type_by_name", "method": "GET", "path": "/types/name/{type_name}"},
    {"module": "alignments", "function": "get_alignment_type", "method": "GET", "path": "/types/{type_id}"},
    {"module": "alignments", "function": "get_alignment", "method": "GET", "path": "/{alignment_id}"},
    {"module": "alignments", "function": "get_alignment_buckets", "method": "GET", "path": "/{alignment_id}/buckets"},
    {"module": "alignments", "function": "get_alignment_buckets_list", "method": "GET", "path": "/{alignment_id}/buckets/list"},
    {"module": "alignments", "function": "get_alignment_events", "method": "GET", "path": "/{alignment_id}/events"},
    {"module": "alignments", "function": "get_alignment_findings", "method": "GET", "path": "/{alignment_id}/findings"},
    {"module": "alignments", "function": "get_alignment_findings_list", "method": "GET", "path": "/{alignment_id}/findings/list"},
    {"module": "alignments", "function": "update_alignment", "method": "PATCH", "path": "/{alignment_id}"},
    {"module": "alignments", "function": "create_alignment", "method": "POST", "path": "/"},
    {"module": "alignments", "function": "execute_alignment_bucket", "method": "POST", "path": "/{alignment_id}/buckets/{bucket_id}/execute"},
    {"module": "alignments", "function": "close_alignment", "method": "POST", "path": "/{alignment_id}/close"},
    {"module": "alignments", "function": "execute_alignment", "method": "POST", "path": "/{alignment_id}/execute"},
    {"module": "alignments", "function": "open_alignment", "method": "POST", "path": "/{alignment_id}/open"},
    {"module": "alignments", "function": "retry_failed_bucket_processing", "method": "POST", "path": "/{alignment_id}/retry-failed-buckets"},
    {"module": "bgtasks", "function": "monitor_task", "method": "GET", "path": "/{job_id}"},
    {"module": "buckets", "function": "delete_bucket", "method": "DELETE", "path": "/{bucket_id}"},
    {"module": "buckets", "function": "get_buckets_by_alignment", "method": "GET", "path": "/alignment/{alignment_id}"},
    {"module": "buckets", "function": "get_buckets_summary_by_alignment", "method": "GET", "path": "/alignment/{alignment_id}/summary"},
    {"module": "buckets", "function": "get_buckets_by_status", "method": "GET", "path": "/status/{status}"},
    {"module": "buckets", "function": "get_buckets_by_type", "method": "GET", "path": "/type/{bucket_type}"},
    {"module": "buckets", "function": "get_bucket", "method": "GET", "path": "/{bucket_id}"},
    {"module": "buckets", "function": "get_bucket_detail", "method": "GET", "path": "/{bucket_id}/detail"},
    {"module": "buckets", "function": "get_bucket_events", "method": "GET", "path": "/{bucket_id}/events"},
    {"module": "buckets", "function": "get_bucket_task_groups_summary", "method": "GET", "path": "/{bucket_id}/task-groups/summary"},
    {"module": "buckets", "function": "update_bucket_status", "method": "PATCH", "path": "/{bucket_id}/status"},
    {"module": "buckets", "function": "create_bucket", "method": "POST", "path": "/"},
    {"module": "commands", "function": "delete_command_group", "method": "DELETE", "path": "/groups/{group_id}"},
    {"module": "commands", "function": "remove_task_association_from_command_group", "method": "DELETE", "path": "/groups/{group_id}/task"},
    {"module": "commands", "function": "delete_command", "method": "DELETE", "path": "/{command_id}"},
    {"module": "commands", "function": "get_command_groups_by_bucket", "method": "GET", "path": "/groups/bucket/{bucket_id}"},
    {"module": "commands", "function": "get_command_groups_by_status", "method": "GET", "path": "/groups/status/{status}"},
    {"module": "commands", "function": "get_command_groups_by_task", "method": "GET", "path": "/groups/task/{task_id}"},
    {"module": "commands", "function": "get_command_group", "method": "GET", "path": "/groups/{group_id}"},
    {"module": "commands", "function": "get_commands_by_group", "method": "GET", "path": "/groups/{group_id}/commands"},
    {"module": "commands", "function": "get_command_group_events", "method": "GET", "path": "/groups/{group_id}/events"},
    {"module": "commands", "function": "get_mappings_by_command", "method": "GET", "path": "/mappings/command/{command_id}"},
    {"module": "commands", "function": "get_mappings_by_finding", "method": "GET", "path": "/mappings/finding/{finding_id}"},
    {"module": "commands", "function": "get_mappings_by_command_group", "method": "GET", "path": "/mappings/group/{group_id}"},
    {"module": "commands", "function": "get_command", "method": "GET", "path": "/{command_id}"},
    {"module": "commands", "function": "get_command_events", "method": "GET", "path": "/{command_id}/events"},
    {"module": "commands", "function": "get_items_created_before_command", "method": "GET", "path": "/{command_id}/items/created-before"},
    {"module": "commands", "function": "get_items_retired_before_command", "method": "GET", "path": "/{command_id}/items/retired-before"},
    {"module": "commands", "function": "get_items_updated_before_command", "method": "GET", "path": "/{command_id}/items/updated-before"},
    {"module": "commands", "function": "query_commands_referencing_command_item", "method": "GET", "path": "/{command_id}/referencing-commands"},
    {"module": "commands", "function": "update_command_group", "method": "PATCH", "path": "/groups/{group_id}"},
    {"module": "commands", "function": "update_command_group_status", "method": "PATCH", "path": "/groups/{group_id}/status"},
    {"module": "commands", "function": "update_command", "method": "PATCH", "path": "/{command_id}"},
    {"module": "commands", "function": "approve_command", "method": "PATCH", "path": "/{command_id}/approve"},
    {"module": "commands", "function": "mark_command_pending", "method": "PATCH", "path": "/{command_id}/pending"},
    {"module": "commands", "function": "reject_command", "method": "PATCH", "path": "/{command_id}/reject"},
    {"module": "commands", "function": "create_command_group", "method": "POST", "path": "/groups"},
    {"module": "commands", "function": "create_command", "method": "POST", "path": "/groups/{group_id}/commands"},
    {"module": "commands", "function": "push_command_group_to_task", "method": "POST", "path": "/groups/{group_id}/task"},
    {"module": "commands", "function": "query_commands_impacting_payload_before_command", "method": "POST", "path": "/impacting-commands"},
    {"module": "commands", "function": "query_display_options_from_items", "method": "POST", "path": "/items/display-options"},
    {"module": "commands", "function": "create_command_findings_mapping", "method": "POST", "path": "/mappings"},
    {"module": "commands", "function": "query_reference_options_before_command", "method": "POST", "path": "/{command_id}/items/reference-options"},
    {"module": "commands", "function": "update_command_findings_mappings", "method": "PUT", "path": "/mappings/command/{command_id}"},
    {"module": "findings", "function": "delete_finding", "method": "DELETE", "path": "/{finding_id}"},
    {"module": "findings", "function": "get_findings_by_alignment", "method": "GET", "path": "/alignment/{alignment_id}"},
    {"module": "findings", "function": "get_findings_by_bucket", "method": "GET", "path": "/bucket/{bucket_id}"},
    {"module": "findings", "function": "list_query_providers", "method": "GET", "path": "/match-queries/providers"},
    {"module": "findings", "function": "get_finding_match_query", "method": "GET", "path": "/match-queries/{match_query_id}"},
    {"module": "findings", "function": "get_finding", "method": "GET", "path": "/{finding_id}"},
    {"module": "findings", "function": "get_finding_events", "method": "GET", "path": "/{finding_id}/events"},
    {"module": "findings", "function": "get_finding_match_queries", "method": "GET", "path": "/{finding_id}/match-queries"},
    {"module": "findings", "function": "update_finding_match_query", "method": "PATCH", "path": "/match-queries/{match_query_id}"},
    {"module": "findings", "function": "update_finding_additional_data", "method": "PATCH", "path": "/{finding_id}/additional-data"},
    {"module": "findings", "function": "move_finding", "method": "PATCH", "path": "/{finding_id}/bucket"},
    {"module": "findings", "function": "create_add_finding", "method": "POST", "path": "/add"},
    {"module": "findings", "function": "create_add_finding_bulk", "method": "POST", "path": "/add/bulk"},
    {"module": "findings", "function": "create_retire_finding", "method": "POST", "path": "/retire"},
    {"module": "findings", "function": "create_retire_finding_list", "method": "POST", "path": "/retire/bulk"},
    {"module": "findings", "function": "create_update_finding", "method": "POST", "path": "/update"},
    {"module": "findings", "function": "create_update_finding_bulk", "method": "POST", "path": "/update/bulk"},
    {"module": "findings", "function": "create_finding_match_query", "method": "POST", "path": "/{finding_id}/match-queries"},
    {"module": "impact_assessments", "function": "get_alignment_dependency_graph", "method": "GET", "path": "/alignment/{alignment_id}/dependency-graph"},
    {"module": "impact_assessments", "function": "get_bucket_compiled_graph", "method": "GET", "path": "/bucket/{bucket_id}/compiled-graph"},
    {"module": "impact_assessments", "function": "get_bucket_dependency_graph", "method": "GET", "path": "/bucket/{bucket_id}/dependency-graph"},
    {"module": "impact_assessments", "function": "get_command_group_graph", "method": "GET", "path": "/command-group/{command_group_id}/command-graph"},
    {"module": "impact_assessments", "function": "get_command_graph", "method": "GET", "path": "/command/{command_id}/command-graph"},
    {"module": "impact_assessments", "function": "get_impact_assessment_by_finding", "method": "GET", "path": "/finding/{finding_id}"},
    {"module": "impact_assessments", "function": "get_finding_dependency_graph", "method": "GET", "path": "/finding/{finding_id}/dependency-graph"},
    {"module": "impact_assessments", "function": "get_finding_dependency_item_type_counts", "method": "GET", "path": "/finding/{finding_id}/dependency-graph/item-type-counts"},
    {"module": "impact_assessments", "function": "get_impact_assessment", "method": "GET", "path": "/{assessment_id}"},
    {"module": "impact_assessments", "function": "expand_alignment_dependency_graph", "method": "POST", "path": "/alignment/{alignment_id}/dependency-graph/expand"},
    {"module": "impact_assessments", "function": "get_bucket_dependency_graph_filtered", "method": "POST", "path": "/bucket/{bucket_id}/dependency-graph"},
    {"module": "impact_assessments", "function": "expand_bucket_dependency_graph", "method": "POST", "path": "/bucket/{bucket_id}/dependency-graph/expand"},
    {"module": "post_processing", "function": "drop_commands_alignment", "method": "DELETE", "path": "/alignments/{alignment_id}/drop-commands"},
    {"module": "post_processing", "function": "drop_commands_bucket", "method": "DELETE", "path": "/buckets/{bucket_id}/drop-commands"},
    {"module": "post_processing", "function": "download_alignment_artifacts_archive", "method": "GET", "path": "/alignments/{alignment_id}/artifacts-archive"},
    {"module": "post_processing", "function": "get_alignment_graph_status", "method": "GET", "path": "/alignments/{alignment_id}/graph/status"},
    {"module": "post_processing", "function": "download_alignment_validation_reports_archive", "method": "GET", "path": "/alignments/{alignment_id}/validation-reports-archive"},
    {"module": "post_processing", "function": "get_post_processing_job", "method": "GET", "path": "/jobs/{job_id}"},
    {"module": "post_processing", "function": "upload_alignment_source_file", "method": "POST", "path": "/alignments/source-file"},
    {"module": "post_processing", "function": "dryrun_commands_alignment", "method": "POST", "path": "/alignments/{alignment_id}/dryrun-commands"},
    {"module": "post_processing", "function": "generate_alignment_findings_summary_report", "method": "POST", "path": "/alignments/{alignment_id}/findings-summary-report"},
    {"module": "post_processing", "function": "register_alignment_graph_roots", "method": "POST", "path": "/alignments/{alignment_id}/graph/register-roots"},
    {"module": "post_processing", "function": "retry_failed_alignment_graph_work_items", "method": "POST", "path": "/alignments/{alignment_id}/graph/retry-failed"},
    {"module": "post_processing", "function": "start_alignment_graph_processing", "method": "POST", "path": "/alignments/{alignment_id}/graph/start-processing"},
    {"module": "post_processing", "function": "map_findings_alignment", "method": "POST", "path": "/alignments/{alignment_id}/map-findings"},
    {"module": "post_processing", "function": "dryrun_commands_bucket", "method": "POST", "path": "/buckets/{bucket_id}/dryrun-commands"},
    {"module": "post_processing", "function": "start_dryrun_commands_bucket_job", "method": "POST", "path": "/buckets/{bucket_id}/dryrun-commands/jobs"},
    {"module": "post_processing", "function": "map_findings_bucket", "method": "POST", "path": "/buckets/{bucket_id}/map-findings"},
    {"module": "post_processing", "function": "refresh_commands_generation_bucket", "method": "POST", "path": "/buckets/{bucket_id}/refresh-commands-generation"},
    {"module": "post_processing", "function": "run_command_generation_bucket", "method": "POST", "path": "/buckets/{bucket_id}/run-command-generation"},
    {"module": "registrations", "function": "delete_processor_registration", "method": "DELETE", "path": "/{registration_id}"},
    {"module": "registrations", "function": "list_processor_registrations", "method": "GET", "path": "/"},
    {"module": "registrations", "function": "get_processor_registration_by_type", "method": "GET", "path": "/by-type/{type_name}"},
    {"module": "registrations", "function": "get_processor_registration", "method": "GET", "path": "/{registration_id}"},
    {"module": "registrations", "function": "update_processor_registration", "method": "PATCH", "path": "/{registration_id}"},
    {"module": "registrations", "function": "upsert_processor_registration", "method": "POST", "path": "/"},
]

__all__ = ["API_ENDPOINTS"]



"""API schemas for dataset processor registrations."""

from __future__ import annotations

import datetime
from uuid import UUID

from app.api.schemas.common import BaseResponse, BaseSchema
from app.core.model import Field


class ProcessorEndpointsSchema(BaseSchema):
    """Routing targets for a registered dataset processor."""

    upload: str | None = Field(
        default=None,
        description="Absolute URL for source-file uploads",
    )
    upload_task_status: str | None = Field(
        default=None,
        description="Template URL for polling upload task status ({task_id})",
    )
    refresh_bucket_commands: str | None = Field(
        default=None,
        description="Template URL for bucket command refresh ({bucket_id})",
    )
    artifacts_download: str | None = Field(
        default=None,
        description="Template URL for artifacts archive download ({alignment_id})",
    )
    validation_reports_download: str | None = Field(
        default=None,
        description="Template URL for validation reports archive download ({alignment_id})",
    )
    retry_failed_buckets: str | None = Field(
        default=None,
        description="Template URL for retrying failed bucket processing ({alignment_id})",
    )


class ProcessorRegistrationBase(BaseSchema):
    """Shared processor registration fields."""

    alignment_type_name: str = Field(
        ...,
        description="Alignment type handled by this processor",
        min_length=1,
        max_length=255,
    )
    processor_name: str = Field(
        ...,
        description="Stable processor identifier",
        min_length=1,
        max_length=255,
    )
    processor_version: str = Field(
        ...,
        description="Deployed processor package version",
        min_length=1,
        max_length=64,
    )
    contract_version: str = Field(
        ...,
        description="alignment-contracts version implemented by this processor",
        min_length=1,
        max_length=32,
    )
    capabilities: list[str] = Field(
        default_factory=list,
        description="Optional capability tags (upload, process, validate, ...)",
    )
    enabled: bool = Field(
        default=True,
        description="Whether routing to this processor is enabled",
    )
    endpoints: ProcessorEndpointsSchema = Field(
        ...,
        description="Routing endpoints for this processor binding",
    )


class ProcessorRegistrationUpsertRequest(ProcessorRegistrationBase):
    """Create or replace the processor binding for an alignment type."""

    pass


class ProcessorRegistrationUpdateRequest(BaseSchema):
    """Partial update for an existing processor registration."""

    processor_name: str | None = Field(default=None, min_length=1, max_length=255)
    processor_version: str | None = Field(default=None, min_length=1, max_length=64)
    contract_version: str | None = Field(default=None, min_length=1, max_length=32)
    capabilities: list[str] | None = None
    enabled: bool | None = None
    endpoints: ProcessorEndpointsSchema | None = None


class ProcessorRegistrationResponse(BaseResponse):
    """Processor registration returned by the API."""

    id: UUID = Field(..., description="Registration identifier")
    alignment_type_id: int = Field(..., description="Bound alignment type identifier")
    alignment_type_name: str = Field(..., description="Bound alignment type name")
    processor_name: str
    processor_version: str
    contract_version: str
    capabilities: list[str]
    enabled: bool
    endpoints: ProcessorEndpointsSchema
    created_at: datetime.datetime
    updated_at: datetime.datetime



from app.core.config import CeleryConfig


def test_celery_config_graph_burst_defaults() -> None:
    cfg = CeleryConfig()

    assert cfg.alignment_graph_enqueue_fetch_limit == 2000
    assert cfg.alignment_graph_processing_max_parallel_tasks == 8
    assert cfg.alignment_graph_retry_base_delay_seconds == 10
    assert cfg.alignment_graph_retry_max_attempts == 5



"""Post-processing response schemas."""

from datetime import datetime
from enum import StrEnum
from typing import Any
from uuid import UUID
from .common import BaseResponse, Field
from ..enums import FindingType


class PostProcessingJobStatus(StrEnum):
    queued = "queued"
    running = "running"
    succeeded = "succeeded"
    failed = "failed"


class PostProcessingJobResponse(BaseResponse):
    job_id: UUID = Field(..., description="ID of the background post-processing job")
    resource_type: str = Field(..., description="Type of resource being processed")
    resource_id: UUID = Field(..., description="ID of the target resource")
    alignment_id: UUID | None = Field(
        default=None,
        description="Owning alignment ID when the job targets a bucket",
    )
    operation: str = Field(..., description="Post-processing operation name")
    status: PostProcessingJobStatus = Field(..., description="Current job status")
    stage: str | None = Field(default=None, description="Current execution stage")
    celery_task_id: str | None = Field(
        default=None,
        description="Underlying Celery task ID, when dispatched",
    )
    result_payload: dict[str, Any] | None = Field(
        default=None,
        description="Job-specific result payload when the job succeeds",
    )
    error_message: str | None = Field(
        default=None,
        description="Terminal error message when the job fails",
    )
    error_kind: str | None = Field(
        default=None,
        description="Normalized terminal error kind when the job fails",
    )
    started_at: datetime | None = Field(
        default=None,
        description="When execution started",
    )
    finished_at: datetime | None = Field(
        default=None,
        description="When execution finished",
    )
    created_at: datetime = Field(..., description="When the job was created")
    updated_at: datetime = Field(..., description="When the job was last updated")


class MapFindingsRequest(BaseResponse):
    keys: list[str] = Field(
        description=(
            "List of finding key fields to use for mapping. If not provided, all "
            "available key fields will be used."
        ),
        min_length=1,
    )
    finding_type: FindingType | None = Field(
        default=None,
        description=(
            "Optional finding type filter to apply during mapping. When specified, "
            "only findings with a matching 'finding_classification.finding_type' "
            "will be processed."
        ),
    )
    filter_item_type_id: UUID | None = Field(
        default=None,
        description=(
            "Optional item type ID filter to apply during mapping. When specified, "
            "only findings with a matching 'item_type_version.item_type_id' will be "
            "processed."
        ),
    )


class AlignmentGraphExpandRequest(BaseResponse):
    work_type: str = Field(
        default="expand_dependencies",
        description="Work type to expand: expand_siblings or expand_dependencies.",
    )
    batch_size: int = Field(
        default=100,
        ge=1,
        le=1000,
        description="Maximum number of work items to claim in this expansion call.",
    )
    worker_id: str | None = Field(
        default=None,
        description="Optional worker identifier used for claim attribution.",
    )


class AlignmentGraphStatusResponse(BaseResponse):
    alignment_id: UUID = Field(..., description="ID of the alignment")
    total: int = Field(..., description="Total work items")
    pending: int = Field(..., description="Pending work items")
    queued: int = Field(..., description="Queued work items")
    retry_wait: int = Field(
        default=0,
        description="Pending work items that are cooling down before retry.",
    )
    in_progress: int = Field(..., description="In-progress work items")
    completed: int = Field(..., description="Completed work items")
    failed: int = Field(..., description="Failed work items")
    last_updated_at: datetime | None = Field(
        default=None,
        description="Most recent UTC timestamp when any graph work item changed",
    )


class AlignmentGraphOperationResponse(BaseResponse):
    alignment_id: UUID = Field(..., description="ID of the alignment")
    items_processed: int = Field(..., description="Number of items processed")


class SiblingKeyConfig(BaseResponse):
    """Per item-type/version sibling expansion key configuration."""

    item_type_id: UUID = Field(..., description="Item type ID")
    version: str | None = Field(
        default=None,
        description="Item type version. When None, the config acts as a wildcard and applies to all versions.",
    )
    keys: list[str] = Field(
        ...,
        min_length=1,
        description="Key fields used to query DSG Core for sibling items.",
    )


class AlignmentGraphStartProcessingRequest(BaseResponse):
    """Optional overrides for starting graph processing."""

    batch_size: int | None = Field(
        default=None,
        ge=1,
        le=1000,
        description="Optional batch size override for this run.",
    )
    timeout_seconds: int | None = Field(
        default=None,
        ge=60,
        le=86400,
        description="Optional timeout override for this run.",
    )
    sibling_keys: list[SiblingKeyConfig] | None = Field(
        default=None,
        description=(
            "Per item-type/version sibling expansion key configs. "
            "When provided, these override the service default keys for each "
            "matching item type (and version) during graph expansion."
        ),
    )


class AlignmentGraphStartProcessingResponse(BaseResponse):
    alignment_id: UUID = Field(..., description="ID of the alignment")
    job_id: str = Field(..., description="Celery job ID handling the graph build")


class BucketFindingsResponse(BaseResponse):
    """Response schema for bucket-scoped findings operations."""

    bucket_id: UUID = Field(..., description="ID of the bucket")
    findings_processed: int = Field(..., description="Number of findings processed")


class BucketCommandsResponse(BaseResponse):
    """Response schema for bucket-scoped commands operations."""

    bucket_id: UUID = Field(..., description="ID of the bucket")
    commands_processed: int = Field(..., description="Number of commands processed")


class AlignmentFindingsResponse(BaseResponse):
    """Response schema for alignment-scoped findings operations."""

    alignment_id: UUID = Field(..., description="ID of the alignment")
    findings_processed: int = Field(..., description="Number of findings processed")


class AlignmentCommandsResponse(BaseResponse):
    """Response schema for alignment-scoped commands operations."""

    alignment_id: UUID = Field(..., description="ID of the alignment")
    commands_processed: int = Field(..., description="Number of commands processed")


class BucketCommandsDropResponse(BaseResponse):
    """Response schema for bucket-scoped command cleanup operations."""

    bucket_id: UUID = Field(..., description="ID of the bucket")
    command_groups_deleted: int = Field(
        ..., description="Number of command groups deleted"
    )
    commands_deleted: int = Field(..., description="Number of commands deleted")
    command_dryruns_deleted: int = Field(
        ..., description="Number of command dry-run records deleted"
    )


class AlignmentCommandsDropResponse(BaseResponse):
    """Response schema for alignment-scoped command cleanup operations."""

    alignment_id: UUID = Field(..., description="ID of the alignment")
    buckets_processed: int = Field(..., description="Number of buckets evaluated")
    command_groups_deleted: int = Field(
        ..., description="Total number of command groups deleted"
    )
    commands_deleted: int = Field(..., description="Total number of commands deleted")
    command_dryruns_deleted: int = Field(
        ..., description="Total number of command dry-run records deleted"
    )


class AlignmentFindingsSummaryReportRequest(BaseResponse):
    """Request schema for findings summary report generation."""

    file_name: str | None = Field(
        default=None,
        description=(
            "Destination object key for the generated Excel file in bucket storage. "
            "When omitted, the API stores the report under "
            "reports/<alignment_name>[_from_<from_version>][_to_<to_version>]_findings_summary.xlsx."
        ),
    )
    sheet_name: str = Field(
        default="findings_summary",
        min_length=1,
        description="Sheet name used inside the generated workbook.",
    )


class AlignmentFindingsSummaryReportResponse(BaseResponse):
    """Response schema for findings summary report generation."""

    alignment_id: UUID = Field(..., description="ID of the alignment")
    output_path: str = Field(
        ..., description="Object key of the generated report in bucket storage"
    )
    sheet_name: str = Field(..., description="Workbook sheet name used for export")
    rows_exported: int = Field(..., description="Number of rows written to the report")
    download_url: str = Field(
        ..., description="Signed URL to download the generated report"
    )


class AlignmentArtifactsArchiveResponse(BaseResponse):
    """Response schema for alignment artifacts archive download URL."""

    alignment_id: UUID = Field(..., description="ID of the alignment")
    download_url: str = Field(
        ..., description="Signed URL to download artifacts archive"
    )



from __future__ import annotations

from contextlib import contextmanager
from unittest.mock import Mock

from app.core import db


def test_ensure_post_processing_active_job_unique_index_skips_when_index_exists(
    monkeypatch,
) -> None:
    connect_cm = Mock()
    connect_cm.__enter__ = Mock(return_value=object())
    connect_cm.__exit__ = Mock(return_value=None)

    monkeypatch.setattr(db.engine, "connect", Mock(return_value=connect_cm))
    monkeypatch.setattr(
        db.engine,
        "begin",
        Mock(side_effect=AssertionError("engine.begin() should not run")),
    )
    monkeypatch.setattr(db, "_index_exists", Mock(return_value=True))

    db._ensure_post_processing_active_job_unique_index()

    db.engine.connect.assert_called_once()
    db.engine.begin.assert_not_called()
    db._index_exists.assert_called_once()


def test_ensure_post_processing_active_job_unique_index_serializes_and_rechecks(
    monkeypatch,
) -> None:
    executed_sql: list[str] = []

    class FakeDuplicateResult:
        def first(self):
            return None

    class FakeScalarResult:
        def scalar(self):
            return None

    class FakeConnection:
        def execute(self, statement, params=None):  # noqa: ANN001
            sql = str(statement)
            executed_sql.append(sql)
            if "GROUP BY resource_type, resource_id, operation" in sql:
                return FakeDuplicateResult()
            return FakeScalarResult()

    @contextmanager
    def fake_connect():
        yield object()

    @contextmanager
    def fake_begin():
        yield FakeConnection()

    index_exists = Mock(side_effect=[False, False])

    monkeypatch.setattr(db.engine, "connect", fake_connect)
    monkeypatch.setattr(db.engine, "begin", fake_begin)
    monkeypatch.setattr(db, "_index_exists", index_exists)

    db._ensure_post_processing_active_job_unique_index()

    assert index_exists.call_count == 2
    assert any("pg_advisory_xact_lock" in sql for sql in executed_sql)
    assert any(
        "CREATE UNIQUE INDEX IF NOT EXISTS uq_post_processing_active_job" in sql
        for sql in executed_sql
    )


def test_ensure_post_processing_active_job_unique_index_skips_create_after_locked_recheck(
    monkeypatch,
) -> None:
    executed_sql: list[str] = []

    class FakeScalarResult:
        def scalar(self):
            return None

    class FakeConnection:
        def execute(self, statement, params=None):  # noqa: ANN001
            executed_sql.append(str(statement))
            return FakeScalarResult()

    @contextmanager
    def fake_connect():
        yield object()

    @contextmanager
    def fake_begin():
        yield FakeConnection()

    index_exists = Mock(side_effect=[False, True])

    monkeypatch.setattr(db.engine, "connect", fake_connect)
    monkeypatch.setattr(db.engine, "begin", fake_begin)
    monkeypatch.setattr(db, "_index_exists", index_exists)

    db._ensure_post_processing_active_job_unique_index()

    assert index_exists.call_count == 2
    assert any("pg_advisory_xact_lock" in sql for sql in executed_sql)
    assert not any(
        "CREATE UNIQUE INDEX IF NOT EXISTS uq_post_processing_active_job" in sql
        for sql in executed_sql
    )
